Con                                    

  _CLKMODE = XTAL1 + PLL16X
  _CLKFREQ = 80_000_000

  ' Settings
  BUFFER_LENGTH = 255           ' Input Buffer Length must fit within input/output 'Index' ranges (currently a byte)
  VERSION       = 10


  ' Bases
  DECIMAL       = 0
  HEX           = 1

  ' Input types
  #0
  NONE
  PULSE
  SERIAL

  ' Sides
  #0
  LEFT
  RIGHT

  ' PWM result offsets
  #0
  HIGH_TIME_CH2
  LOW_TIME_CH2
  HIGH_TIME_CH1
  LOW_TIME_CH1

  ' PWM times
  PWM_MIN         = _CLKFREQ / 1_000_000 * 500          ' 500 microseconds
  PWM_MEDIAN      = _CLKFREQ / 1_000_000 * 1_500        ' 1,500 microseconds
  PWM_MAX         = _CLKFREQ / 1_000_000 * 2_500        ' 2,500 microseconds

  ' Motor control modes
  #0
  POWER
  VELOCITY
  STOPPING
  POSITION

  ' Pin assignments
  ' Encoders
  L_ENCODERS_PIN = 0
  R_ENCODERS_PIN = 18
  ' Motors
  L_FAULT_LED   = 2   
  R_FAULT_LED   = 17

  ' Inputs
  CH1           = 21
  CH2           = 20
  AUX1          = 22
  AUX2          = 23

  FAN           = 27

  ' I2C
  SCL           = 28
  SDA           = 29
  ' RS-232 TTL

  ' Terminal Settings
  PROCESSRATE   = _clkfreq / 16_000
  TIMEOUT       = 10

  ' ASCII commands
  NUL           = $00           ' Null character
  SOH           = $01           ' Start of Header
  STX           = $02           ' Start of Text
  ETX           = $03           ' End of Text
  EOT           = $04           ' End of Transmission
  ENQ           = $05           ' Enquiry
  ACK           = $06           ' Acknowledgment
  BEL           = $07           ' Bell
  BS            = $08           ' Backspace
  HT            = $09           ' Horizontal Tab
  LF            = $0A           ' Line feed
  VT            = $0B           ' Vertical Tab
  FF            = $0C           ' Form feed
  CR            = $0D           ' Carriage return
  SO            = $0E           ' Shift Out
  SI            = $0F           ' Shift In
  DLE           = $10           ' Data Link Escape
  DC1           = $11           ' Device Control 1 (i.e. XON)
  DC2           = $12           ' Device Control 2
  DC3           = $13           ' Device Control 3 (i.e. XOFF)
  DC4           = $14           ' Device Control 4
  NAK           = $15           ' Negative Acknowledgment
  SYN           = $16           ' Synchronous idle
  ETB           = $17           ' End of Transmission Block
  CAN           = $18           ' Cancel
  EM            = $19           ' End of Medium
  SB            = $1A           ' Substitute
  ESC           = $1B           ' Escape
  FS            = $1C           ' File Separator
  GS            = $1D           ' Group Separator
  RS            = $1E           ' Record Separator
  US            = $1F           ' Unit Separator
  DEL           = $7F           ' Delete

  ' EEPROM constants
  I2C_ACK       = 0
  I2C_NACK      = 1
  DEVICE_CODE   = %0110 << 4
  PAGE_SIZE     = 128

  ' Positions one wheel must travel, while other wheel is still, to rotate the entire robot 360°
  DEFAULT_POSITIONS_PER_ROTATION = 744

  ' Other
  DIVISOR       = 50            ' Iterations per second
  MILISECOND    = _clkfreq / 1000
  HALF_SEC      = DIVISOR / 2   ' Iterations per unit
  DEFAULT_DEADZONE = 1
  SAFEZONE      = 10
  PROMPT_SUM    = $0
  DEFAULT_PACE  = 0
  DEFAULT_HOLD  = 250
  DEFAULT_PULSE_SCALE = _CLKFREQ / 1_000_000 * 500
  DEFAULT_KP    = 100
  DEFAULT_KI    = 65
  DEFAULT_KI_MAX = Motors#MAX_ON_TIME * 10 / 100
  DEFAULT_KI_TIME_DAMPEN = 65
  DEFAULT_KI_LIVE_ZONE = 8
  DEFAULT_KI_LIMIT_BY_POWER = 512
  DEFAULT_BAUDMODE = %1100
  DEFAULT_TX = CH1
  DEFAULT_RX = CH1
  DEFAULT_BAUDRATE = 19_200
  DEFAULT_MAX_POS_ACC = 512 / 4
  DEFAULT_MAX_POW_ACC = Motors#MAX_ON_TIME / 16
  USB_TX = 30
  USB_RX = 31
  EXPECTED_HW_VER = $0001
  NUM_HB_FAULTS_TO_STOP = 3 ' Number of consecutive H Bridge faults required to stop control

           
Obj                             
                                
  Encoders[2]   : "Quadrature_Encoder"
  Term          : "FullDuplexSerial"
  PWM           : "ReadPulseWidths"
  Motors        : "DHB-10 Motor Driver"


Var

  long MotorSpeed[2]                                    ' Current motor speeds reported from position controllers  
  long Stack[50 * 2]                                    ' Stack for cog running position control system
  long GlobalParameter


Dat

PWMResults    long      0[2 * 2]                        ' Results from cog recording PWM pulses

PrevMotPoss   long      0[HALF_SEC * 2]                 ' Previous motor possitions
MotorPosition long      0[2]                            ' Physical position of the motor as returned by encoder
MotPosOffset  long      0[2]                            ' Offset between Physical and internal motor position
MotorPower    long      0[2]                            ' Current motor power level
MidPosition   long      0[2]                            ' Position the PD loop is attempting to hold
MidPosAcc     long      0[2]                            '  fractional part
MidVelocity   long      0[2]                            ' Current position change velocity
MidVelAcc     long      0[2]                            '  fractional part
SetPosition   long      0[2]                            ' Position that MidPosition is approaching
SetVelocity   long      0[2]                            ' Velocity that MidVelocity is approaching
SetPower      long      0[2]                            ' Power level that MotorPower is approaching
MaxPowAccel   long      DEFAULT_MAX_POW_ACC             ' Maximum allowed motor power acceleration
MaxPosAccel   long      DEFAULT_MAX_POS_ACC             ' Maximum allowed positional acceleration
Decel         long      0[2]                            ' Deceleration rate
StartPosition long      0[2]
Kp            long      DEFAULT_KP                      ' Proportional error feedback gain
KiMax         long      DEFAULT_KI_MAX
KiAcc         long      0[2]
KiTimeDampen  long      DEFAULT_KI_TIME_DAMPEN
Baudrate      long      DEFAULT_BAUDRATE
Hold          long      DEFAULT_HOLD
KiLimitByPwr  long      DEFAULT_KI_LIMIT_BY_POWER
PDRunning     long      -1
ProcessPWMCog long      -1
PulseScale    long      DEFAULT_PULSE_SCALE
PosPerRot     long      DEFAULT_POSITIONS_PER_ROTATION
HWVer         long      65534
Ki            long      DEFAULT_KI                      ' Integral error feedback gain
InsideWheel   byte      0
KiLiveZone    byte      DEFAULT_KI_LIVE_ZONE
Pace          byte      DEFAULT_PACE
Deadzone      byte      DEFAULT_DEADZONE
Baudmode      byte      DEFAULT_BAUDMODE
Tx            byte      DEFAULT_TX
Rx            byte      DEFAULT_RX

StillCnt      byte      0[2]                            ' Number of iterations in a row that the motor hasn't moved noticeably
Mode          byte      POWER                           ' Current mode of the control system

InputBuffer   byte      0[BUFFER_LENGTH]                ' Command input buffer
OutputBuffer  byte      0[BUFFER_LENGTH]                ' Response output buffer
InputIndex    byte      0                               ' Next empty position in the command input buffer
ParseIndex    byte      0                               ' Last processed position in the command input buffer 
PrvMotPossIdx byte      0                               ' Previous motor possitions index
OutputIndex   byte      0                               ' Next empty position in the response output buffer
Verbose       byte      True                            ' Verbosity level (Currently nonzero = verbose)
KeepAlive     byte      False                           ' State of keep alive requirement
Echo          byte      False                           ' State of terminal loopback
Base          byte      DECIMAL                         ' Numeric base for serial commands
RightFaultCnt    byte 0
LeftFaultCnt    byte 0


Pub Main

  Init
  ProcessPWMCog := cognew(ProcessPWM, @Stack[50])  
  ProcessSerialInput


Pub Init

  CheckShortedAUX                                      ' Reset defaults if AUX 1 & 2 are shorted
  GetHWVer
  PDRunning := cognew(PDLoop, @Stack)                   ' Run the position controller in another core


Pub GetHWVer

  outa[SCL]~
  outa[SDA]~
  dira[SCL]~                    ' Float clock high
  dira[SCL]~~                   ' Output clock low
  dira[SDA]~~                   ' Output data low 
  dira[SCL]~                    ' Float clock high
  dira[SDA]~~                   ' Output data low
  dira[SDA]~                    ' Float data high
  
  dira[FAN] := outa[FAN]~~                              ' Keep fan on

  ReadEEPROM(@HWVer, @HWVer + 1, HWVer)       ' Read the hardware version from upper EEPROM


Pub Stop

  Motors.Stop
  Term.Stop
  PWM.Stop
  Encoders[LEFT].Stop                  
  Encoders[RIGHT].Stop
  CogStop(ProcessPWMCog)
  CogStop(PDRunning)


Pub ProcessSerialInput | RxCheck, NextCNT, LoopCount, Error

  Term.Start(Rx, Tx, Baudmode, Baudrate)                ' Start a UART for the command port

  repeat                                                ' Main loop (repeats forever)
    NextCNT := PROCESSRATE + cnt
    repeat while (RxCheck := Term.RxCheck) < 0          ' Read a byte from the command UART
      LoopCount++
      if KeepAlive and LoopCount > constant(_clkfreq / PROCESSRATE)' Stop motors if no com in one second
        waitcnt(constant(_clkfreq / DIVISOR) + cnt)
        SetPower[LEFT]~
        SetPower[RIGHT]~
        StillCnt[LEFT]~ 
        StillCnt[RIGHT]~
        Mode := POWER
      else
        waitcnt(NextCNT += PROCESSRATE)
    ifnot RxCheck & %1000_0000 or RxCheck == 0
      if Echo
        Term.tx(RxCheck)
      InputBuffer[InputIndex++] := RxCheck
      if ProcessPWMCog <> -1
        cogstop(ProcessPWMCog)
        PWM.Stop
        SetPower[LEFT]~
        SetPower[RIGHT]~
        ProcessPWMCog := -1
      LoopCount~
      if InputIndex == BUFFER_LENGTH                      ' Check for a full buffer
        if Verbose                                        ' If in verbose mode, add a description
          OutputStr(@Nack)                                  ' Ready error response
          OutputStr(@Overflow + 2)
        else
          OutputStr(@Overflow)
          OutputStr(String(" E"))
        repeat                                            ' Ignore all inputs other than CR (terminating a command)        
          if Term.Rx == CR                                ' Send error response
            SendResponse
            quit
      else                                                ' If there isn't a buffer overflow...                              
        case InputBuffer[InputIndex - 1]                  ' Parse the character
          
          BS :                                            ' Process backspaces
            if --InputIndex                               ' Ignore the BS character itself
              --InputIndex                                ' Ignore previous character if exists
       
          CR:                                             ' End command in plaintext mode:
            if InputIndex > 1                             '   Only parse buffer if it has content
              if Error := \Parse                          '   Run the parser and trap and report errors
                OutputIndex~                              '    Handle errors, if they occurred
                if Verbose
                  OutputStr(@Nack)
                  OutputStr(Error + 2)
                else
                  OutputStr(Error)
                  OutputStr(String(" E"))
              SendResponse                                '   Send a response if no error
            else                                          '   For an empty buffer, clear the pointer
              InputIndex~                                 '   to start receiving a new command
       
          SOH..BEL, LF..FF, SO..US, ",", 127..255 :       ' Ignore invalid characters


Pri ProcessPWM | LastPulseCH1, LastPulseCH2, LastCNT, CurrentPhsa, CurrentPhsb, LastPhsa, LastPhsb, PulseTimeCH1, PulseTimeCH2

  frqa := 1                                                                     ' Use ctra to track left-motor pulses received
  ctra := constant(%01110 << 26 | CH1)                                          '  by adding one to phsa on each pulse, and
  phsa~                                                                         '  clearing phsa after processing it
  PulseTimeCH1 := Hold
  outa[L_FAULT_LED]~~

  frqb := 1                                                                     '  Do the same with ctrb for right-motor pulses
  ctrb := constant(%01110 << 26 | CH2)
  phsb~
  PulseTimeCH2 := Hold
  outa[R_FAULT_LED]~~

  PWMResults := |< CH1 | |< CH2
  PWM.Start(@PWMResults)                                              ' Start the PWM reading cog
  Mode := POWER
  SetPower[LEFT]~                                                               ' Turn of the motors
  SetPower[RIGHT]~
  
  LastCNT := cnt
  repeat
    waitcnt(LastCNT += MILISECOND)                                              
    LastPulseCH1 := PWMResults[HIGH_TIME_CH1]                                   ' Read pulse widths only once periteration
    LastPulseCH2 := PWMResults[HIGH_TIME_CH2]

    dira[L_FAULT_LED]~
    CurrentPhsa := phsa
    if LastPulseCH1 => PWM_MIN and LastPulseCH1 =< PWM_MAX and CurrentPhsa <> LastPhsa ' If there is a new left-motor pulse in range
      Mode := POWER                                                             '   Switch to power mode
      SetPower[LEFT]  := (LastPulseCH1 - PWM_MEDIAN) * Motors#MAX_ON_TIME / PulseScale ' Convert the pulse width to a power level
      PulseTimeCH1~                                                             '   Clear the time since last valid pulse
    else                                                                        ' If there is no new pulse or it is out of range
      if PulseTimeCH1++ => Hold and Hold                                        ' and if it has been too long since the last pulse
        SetPower[LEFT]~                                                         '   and the last valid input was a pulse, then turn the motor off
        dira[L_FAULT_LED] := cnt >> 26
      PulseTimeCH1 <#= Hold                                                     '   Prevent PulseTimeCH1 from overflowing
    LastPhsa := CurrentPhsa

    dira[R_FAULT_LED]~
    CurrentPhsb := phsb
    if LastPulseCH2 => PWM_MIN and LastPulseCH2 =< PWM_MAX and CurrentPhsb <> LastPhsb ' Do the same for the right motor
      Mode := POWER                                                                    
      SetPower[RIGHT] := (LastPulseCH2 - PWM_MEDIAN) * Motors#MAX_ON_TIME / -PulseScale
      PulseTimeCH2~
    else
      if PulseTimeCH2++ => Hold and Hold
        SetPower[RIGHT]~
        dira[R_FAULT_LED] := cnt >> 26
      PulseTimeCH2 <#= Hold
    LastPhsb := CurrentPhsb
         
  
Pub Travel(Distance, Speed, Rotation)

  CheckHWVer
  if PDRunning < 1                                    ' Return an error if the position controller cog isn't running
    return @MotorError
  case Speed
    2..32767:
    other:                                            ' Speed must be positive
      return @InvalidParameter      
  case Rotation
    -32767..-1:                                       ' Set which wheel is inside of arc
      InsideWheel := LEFT
    0..32767:                                         ' Default to right when angle = 0
      InsideWheel := RIGHT
  case Distance
    -32767..-1:
      -Rotation                                      ' Negate angle if traveling in reverse
    0:
      ifnot Rotation
        return @InvalidParameter      
    1..32767:
    other:
      return @InvalidParameter      
  if Mode <> VELOCITY
    InterpolateMidVariables
  Mode := POWER                                       ' Stop PDIteration from modifying mid-positions
  KiAcc[LEFT] := KiAcc[RIGHT]~~  
  StartPosition[LEFT] := MidPosition[LEFT]
  StartPosition[RIGHT] := MidPosition[RIGHT]
  SetVelocity[LEFT] := SetVelocity[RIGHT] := Speed / 2 ' Divide by two to convert pos/sec to pos/half sec
  SetPosition[LEFT] := MidPosition[LEFT] + Distance + Rotation * (PosPerRot / 2) / 360
  SetPosition[RIGHT] := MidPosition[RIGHT] + Distance - Rotation * (PosPerRot / 2) / 360
  StillCnt[LEFT]~     
  StillCnt[RIGHT]~
  Mode := POSITION


Pub Move(LeftDistance, RightDistance, Speed)

  CheckHWVer
  if PDRunning < 1                                  ' Return an error if the position controller cog isn't running
    return @MotorError
  case LeftDistance
    -32767..32767:
    other:
      return @InvalidParameter      
  case RightDistance
    -32767..32767:
    other:
      return @InvalidParameter      
  case Speed                                            ' Speed must be positive
    2..32767:
    other:
      return @InvalidParameter      
  if ||LeftDistance < ||RightDistance                 ' Set which wheel is inside of arc
    InsideWheel := LEFT
  else                                                ' Default to right when angle = 0
    InsideWheel := RIGHT
  if Mode <> VELOCITY
    InterpolateMidVariables
  Mode := POWER                                       ' Stop PDIteration from modifying mid-positions 

  'Term.Str(String(CR, LF, "OldLeft")

  KiAcc[LEFT] := KiAcc[RIGHT]~~  
  StartPosition[LEFT] := MidPosition[LEFT]
  StartPosition[RIGHT] := MidPosition[RIGHT]
  SetVelocity[LEFT] := SetVelocity[RIGHT] := Speed / 2 ' Divide by two to convert pos/sec to pos/half sec
  SetPosition[LEFT] := MidPosition[LEFT] + LeftDistance
  SetPosition[RIGHT] := MidPosition[RIGHT] + RightDistance
  StillCnt[LEFT]~     
  StillCnt[RIGHT]~
  Mode := POSITION


Pub GoSpeed(LeftVelocity, RightVelocity)

  LeftVelocity /= 2         ' Divide by two to convert pos/sec to pos/half sec
  RightVelocity /= 2         ' Divide by two to convert pos/sec to pos/half sec
  CheckHWVer
  if PDRunning < 1                                  '   Requires use of the PD controller
    return @MotorError                                 '   Abort if it isn't running
  case LeftVelocity
    -32767..32767:
    other:
      return @InvalidParameter      
  case RightVelocity
    -32767..32767:
    other:
      return @InvalidParameter      
  if Mode <> VELOCITY
    InterpolateMidVariables
  Mode := POWER                                   ' Stop PDIteration from modifying mid-positions 
  KiAcc[LEFT]~~
  KiAcc[RIGHT]~~
  SetVelocity[LEFT] := LeftVelocity                   ' Then set the motors' powers
  SetVelocity[RIGHT] := RightVelocity
  StillCnt[LEFT]~ 
  StillCnt[RIGHT]~
  Mode := VELOCITY


Pub Go(LeftPower, RightPower)

  CheckHWVer
  if PDRunning < 1                                  ' Requires use of the PD controller
    return @MotorError                                 ' Abort if it isn't running
  case LeftPower
    -127..127:
    other:
      return @InvalidParameter      
  case RightPower
    -127..127:
    other:
      return @InvalidParameter      
  Mode := POWER
  SetPower[LEFT] := LeftPower * Motors#MAX_ON_TIME / 127' Then set the motors' powers
  SetPower[RIGHT] := RightPower * Motors#MAX_ON_TIME / 127


Pub Arc(Radius, Speed, Rotation)    

  CheckHWVer
  if PDRunning < 1                                  ' Return an error if the position controller cog isn't running
    return @MotorError
  case Speed
    2..32767:
    other:                                            ' Speed must be positive
      return @InvalidParameter      
  case Rotation                                 ' Set wheel in inside of arc  
    -32767..-1:
    1..32767:                                       ' Default to right when angle = 0
    other:
      return @InvalidParameter    
  case Radius
    0..32767:
      InsideWheel := RIGHT
    -32767..-1:
      InsideWheel := LEFT
      -Rotation                                   ' Negate angle if traveling in reverse
    other:
      return @InvalidParameter
  ' Distance = 2 * pi * radius * degrees / 360        ' 115 ~ 360 / pi, gives us distance along arc
  Radius := Radius * Rotation * 2 / 115
  if Mode <> VELOCITY
    InterpolateMidVariables
  Mode := POWER                                       ' Stop PDIteration from modifying mid-positions
  KiAcc[LEFT] := KiAcc[RIGHT]~~  
  StartPosition[LEFT] := MidPosition[LEFT]
  StartPosition[RIGHT] := MidPosition[RIGHT]
  SetVelocity[LEFT] := SetVelocity[RIGHT] := Speed / 2 ' Divide by two to convert pos/sec to pos/half sec
  SetPosition[LEFT] := MidPosition[LEFT] + Radius + Rotation * (PosPerRot / 2) / 360
  SetPosition[RIGHT] := MidPosition[RIGHT] + Radius - Rotation * (PosPerRot / 2) / 360
  StillCnt[LEFT]~   
  StillCnt[RIGHT]~
  Mode := POSITION


Pub StopIn(Distance)

  if PDRunning < 1                                  ' If the motor controller isn't running
    Return @MotorError
  if Distance
    if Mode <> VELOCITY
      InterpolateMidVariables
    Mode := POWER                                     ' Stop PDIteration from modifying mid-positions
    KiAcc[LEFT] := KiAcc[RIGHT]~~  
    case SetPower[LEFT]
      -1..negx:
        SetPosition[LEFT] := MotorPosition[LEFT] - Distance
      1..posx:
        SetPosition[LEFT] := MotorPosition[LEFT] + Distance
      other:
        SetPosition[LEFT] := MotorPosition[LEFT]
    case SetPower[RIGHT]
      -1..negx:
        SetPosition[RIGHT] := MotorPosition[RIGHT] - Distance
      1..posx:
        SetPosition[RIGHT] := MotorPosition[RIGHT] + Distance
      other:
        SetPosition[RIGHT] := MotorPosition[RIGHT]
    Decel[LEFT] := ||(-HALF_SEC * MidVelocity[LEFT] * MidVelocity[LEFT] / (MidVelocity[LEFT] - HALF_SEC * ||(SetPosition[LEFT] - MotorPosition[LEFT])))
    Decel[RIGHT] := ||(-HALF_SEC * MidVelocity[RIGHT] * MidVelocity[RIGHT] / (MidVelocity[RIGHT] - HALF_SEC * ||(SetPosition[RIGHT] - MotorPosition[RIGHT])))
    StillCnt[LEFT]~ 
    StillCnt[RIGHT]~
    Mode := STOPPING
  else                                              '   For a zero stopping distance:
    SetPower[LEFT]~                                 '     Disable power to both motors
    SetPower[RIGHT]~
    StillCnt[LEFT]~ 
    StillCnt[RIGHT]~
    Mode := POWER                                   '     Set the motors to power mode


Pub Turn(Rotation, Speed)  

  CheckHWVer
  if PDRunning < 1
    return @MotorError
  case Rotation
    -32767..-1:
    1..32767:
    other:
      return @InvalidParameter      
  case Speed
    2..32767:                                   
      if Mode <> VELOCITY
        InterpolateMidVariables
      Mode := POWER                                   ' Stop PDIteration from modifying mid-positions 
      KiAcc[LEFT] := KiAcc[RIGHT]~~  
      StartPosition[LEFT] := MidPosition[LEFT]
      StartPosition[RIGHT] := MidPosition[RIGHT]
      SetVelocity[LEFT] := SetVelocity[RIGHT] := Speed / 2 ' Divide by two to convert pos/sec to pos/half sec
      SetPosition[LEFT] := MidPosition[LEFT] + Rotation ' Add travel distance to current position
      SetPosition[RIGHT] := MidPosition[RIGHT] - Rotation
      StillCnt[LEFT]~ 
      StillCnt[RIGHT]~
      Mode := POSITION
    other:
      return @InvalidParameter


Pub Acceleration(Rate)

  case Rate
    1..32767:                                   
      MaxPosAccel := Rate / 4
    other:
      return @InvalidParameter


Pub Ramp(Rate)

  case Rate
    20..32767:                                   
      MaxPowAccel := Motors#MAX_ON_TIME * 20 / Rate
    other:
      return @InvalidParameter


Pri Parse | Index, Parameter[3], Response               '' Parse the command in the input buffer
  
  InputBuffer[InputIndex - 1]~                          ' Ensure buffer is NUL terminated (may have been CR terminated)

  repeat ParseIndex from 0 to InputIndex - 2            ' Find the end of the command in the input buffer             
    case InputBuffer[ParseIndex]                         
      "a".."z" :                                        ' Set all command characters to uppercase
        InputBuffer[ParseIndex] -= constant("a" - "A")
      HT, " " :                                         ' Determine command length by finding the first whitespace character
        InputBuffer[ParseIndex]~                        ' Null terminate the command
        quit                                            ' ParseIndex now points to the null at the end of the command string (before the first parameter, if present)

  ' Parameter[n] := ParseNum(NextParameter) caches the parameters to check for their existence before running the command
  ' CheckLastParameter checks for too many parameters
  ' Output...() writes responses to the output buffer

  ' Check command against the following strings:
  if     strcomp(@InputBuffer, string("HWVER"))         ' 1 Command: Respond with hardware version number
    CheckLastParameter
    OutputNum(HWVer, 4)
  elseif strcomp(@InputBuffer, string("VER"))           ' 1 Command: Respond with version number
    CheckLastParameter
    OutputNum(VERSION, 4)
  elseif strcomp(@InputBuffer, string("TRVL"))          ' 2 Command: Travel a specified distance at a specified speed
    Parameter[0] := ParseNum(NextParameter)   ' Dist                                                                    
    Parameter[1] := ParseNum(NextParameter)   ' Speed
    if Parameter[2] := OptionalParameter      ' Deg = 0 if no parameter
      Parameter[2] := ParseNum(Parameter[2])  ' otherwise it is a pointer that needs parsingtrvl 
      CheckLastParameter
    if Response := Travel(Parameter[0], Parameter[1], Parameter[2])
      abort Response
  elseif strcomp(@InputBuffer, string("MOVE"))          ' 3 Command: Travel a specified distance at a specified speed
    Parameter[0] := ParseNum(NextParameter) ' L Dist                                                                    
    Parameter[1] := ParseNum(NextParameter) ' R Dist
    Parameter[2] := ParseNum(NextParameter) ' Speed
    CheckLastParameter
    if Response := Move(Parameter[0], Parameter[1], Parameter[2])
      abort Response
  elseif strcomp(@InputBuffer, string("GOSPD"))         ' 4 Command: Set the motors to a specified speed
    Parameter[0] := ParseNum(NextParameter)
    Parameter[1] := ParseNum(NextParameter)
    CheckLastParameter
    if Response := GoSpeed(Parameter[0], Parameter[1])
      abort Response
  elseif     strcomp(@InputBuffer, string("GO"))        ' 5 Command: Set the motors to a specified power
    Parameter[0] := ParseNum(NextParameter)             ' Read both parameters before processing them
    Parameter[1] := ParseNum(NextParameter)             ' To prevent changes with a "Too few parameters" error
    CheckLastParameter
    if Response := Go(Parameter[0], Parameter[1])
      abort Response
  elseif strcomp(@InputBuffer, string("ARC"))           ' 6 Command: Slow to a stop over a specified distance
    Parameter[0] := ParseNum(NextParameter) ' Radius                                                                    
    Parameter[1] := ParseNum(NextParameter) ' Speed
    Parameter[2] := ParseNum(NextParameter) ' Deg
    CheckLastParameter
    if Response := Arc(Parameter[0], Parameter[1], Parameter[2])
      abort Response
{
  elseif strcomp(@InputBuffer, string("STOP"))          ' 7 Command: Slow to a stop over a specified distance
    Parameter := ParseNum(NextParameter)
    CheckLastParameter
    if Response := StopIn(Parameter)
      abort Response
    CheckHWVer
}
  elseif strcomp(@InputBuffer, string("TURN"))        ' 8 Command: Turn in place a specified number of degrees
    ' Read both parameters before processing them (and adjust encoder positions per revolution / 2 wheels)
    Parameter[0] := ParseNum(NextParameter) ' angle
    Parameter[1] := ParseNum(NextParameter) ' speed
    CheckLastParameter
    if Response := Turn(Parameter[0], Parameter[1])
      abort Response
  elseif strcomp(@InputBuffer, string("ACC"))           ' 9
    if Parameter := OptionalParameter
      Parameter := ParseNum(Parameter)
      CheckLastParameter
      if Response := Acceleration(Parameter[0])
        abort Response
    else
      OutputNum(MaxPosAccel * 4, 4)
  elseif strcomp(@InputBuffer, string("RAMP"))          ' 10
    if Parameter := OptionalParameter
      Parameter := ParseNum(Parameter)
      CheckLastParameter
      if Response := Ramp(Parameter[0])
        abort Response
    else
      OutputNum(Motors#MAX_ON_TIME * 20 / MaxPowAccel, 4)
  elseif     strcomp(@InputBuffer, string("SPD"))       ' 11 Command: Respond with the current motor speeds
    CheckLastParameter
    OutputNum(MotorSpeed[LEFT] * 2, 4)                  ' Multiply by two to convert pos/half sec to pos/sec
    OutputChr(" ")
    OutputNum(MotorSpeed[RIGHT] * 2, 4)
  elseif strcomp(@InputBuffer, string("HEAD"))          ' 12 Command: Respond with the current heading, relative to last reset
    CheckLastParameter
    OutputNum((((MotorPosition[LEFT] + MotPosOffset[LEFT]) - (MotorPosition[RIGHT] + MotPosOffset[RIGHT])) // PosPerRot) * 360 / PosPerRot, 3)
  elseif strcomp(@InputBuffer, string("DIST"))          ' 13 Command: Respond with the accumulative distance each motor has traveled, relative to last reset
    CheckLastParameter
    OutputNum(MotorPosition[LEFT] + MotPosOffset[LEFT], 8)
    OutputChr(" ")
    OutputNum(MotorPosition[RIGHT] + MotPosOffset[RIGHT], 8)
  elseif strcomp(@InputBuffer, string("RST"))           ' 14 Command: Reset heading and accumulated distance
    CheckLastParameter
    MotPosOffset[LEFT] := - MotorPosition[LEFT]
    MotPosOffset[RIGHT] := - MotorPosition[RIGHT]
  elseif strcomp(@InputBuffer, string("VERB"))          ' 15 Command: Set verbose mode
    Parameter := NextParameter
    CheckLastParameter
    if       strcomp(Parameter, string("ON")) | strcomp(Parameter, string("1"))
      Verbose := 1
    elseif strcomp(Parameter, string("OFF")) | strcomp(Parameter, string("0"))
      Verbose := 0
    else
      abort @InvalidParameter        
  else                                                  ' Only 16 elseifs, nest within an else for more
    if strcomp(@InputBuffer, string("ECHO"))            ' Command: Set echo mode
      Parameter := NextParameter
      CheckLastParameter
      if       strcomp(Parameter, string("ON")) | strcomp(Parameter, string("1"))
        if Rx == Tx
          abort @InvalidParameter
        Echo := 1
      elseif strcomp(Parameter, string("OFF")) | strcomp(Parameter, string("0"))
        Echo := 0
      else
        abort @InvalidParameter        
    elseif strcomp(@InputBuffer, string("HEX"))         ' 1 Command: Set base to Hexadecimal
      CheckLastParameter
      Base := HEX
    elseif strcomp(@InputBuffer, string("DEC"))         ' 2 Command: Set base to Decimal
      CheckLastParameter
      Base := DECIMAL
    elseif strcomp(@InputBuffer, string("SETLF"))       ' 3 Command: Respond with CR LF instead of just CR
      Parameter := ParseNum(NextParameter)
      CheckLastParameter
      case Parameter
        0:
          Prompt[1]~
        1:
          Prompt[1] := LF
        other:
          abort @InvalidParameter
    elseif strcomp(@InputBuffer, string("HOLD"))        ' 4 Command: Set hold time after last pulse, in ms
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          0, 20..30_000:                                ' 20 ms to 30 seconds
            Hold := Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(Hold, 4)
    elseif strcomp(@InputBuffer, string("PULSE"))       ' 5 Command: Switch to pulse input mode
      CheckLastParameter
      ProcessPWMCog := cognew(ProcessPWM, @Stack[50])   ' Run the PWM input processer in another core  
    elseif strcomp(@InputBuffer, string("SCALE"))       ' 6 Command: Set pulse scale, in us from center (center is 1,500 us)
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          100..1000:
            PulseScale := constant(_CLKFREQ / 1_000_000) * Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(PulseScale / constant(_CLKFREQ / 1_000_000), 4)
    elseif strcomp(@InputBuffer, string("KP"))          ' 7 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          1..32767:
            Kp := Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(Kp, 4)
    elseif strcomp(@InputBuffer, string("KIMAX"))       ' 8 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          0..100:
            KiMax := Motors#MAX_ON_TIME * Parameter / 100
          other:
            abort @InvalidParameter
      else
        OutputNum(KiMax * 100 / Motors#MAX_ON_TIME, 4)
    elseif strcomp(@InputBuffer, string("BAUD"))        ' 9 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          19_200..115_200:
            Baudrate := Parameter
          other:
            abort @InvalidParameter
        Term.Start(Rx, Tx, Baudmode, Baudrate)                
      else
        OutputNum(Baudrate, 4)
    elseif strcomp(@InputBuffer, string("TXPIN"))       ' 10 Command: Set
      Parameter := NextParameter
      CheckLastParameter
      CheckHWVer
      if     strcomp(Parameter, string("CH1"))
        Tx := CH1
        Baudmode |= %0100
      elseif strcomp(Parameter, string("CH2"))  
        Tx := CH2
        Baudmode |= %0100
      elseif strcomp(Parameter, string("PROG"))  
        Tx := USB_TX
        Baudmode &= %1011
      else
        abort @InvalidParameter      
      if Rx == Tx
        Baudmode |= %1000
        Echo := 0
      else
        Baudmode &= %0111
      Term.Start(Rx, Tx, Baudmode, Baudrate)                
    elseif strcomp(@InputBuffer, string("RXPIN"))       ' 11 Command: Set
      Parameter := NextParameter
      CheckLastParameter
      CheckHWVer
      if     strcomp(Parameter, string("CH1"))
        Rx := CH1
      elseif strcomp(Parameter, string("CH2"))  
        Rx := CH2
      elseif strcomp(Parameter, string("PROG"))  
        Rx := USB_RX
      else
        abort @InvalidParameter      
      if Rx == Tx
        Baudmode |= %1000
        Echo := 0
      else
        Baudmode &= %0111
      Term.Start(Rx, Tx, Baudmode, Baudrate)                
    elseif strcomp(@InputBuffer, string("KIT"))         ' 12 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          0..32767:
            KiTimeDampen := Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(KiTimeDampen, 4)
    elseif strcomp(@InputBuffer, string("KIP"))         ' 13 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          0..32767:
            KiLimitByPwr := Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(KiLimitByPwr, 4)
    elseif strcomp(@InputBuffer, string("LZ"))          ' 14 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          0..255:
            KiLiveZone := Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(KiLiveZone, 4)
    elseif strcomp(@InputBuffer, string("DZ"))          ' 15 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          0..255:
            Deadzone := Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(Deadzone, 4)
    elseif strcomp(@InputBuffer, string("PPR"))         ' 16 Command: Set
      if Parameter := OptionalParameter
        Parameter := ParseNum(Parameter)
        CheckLastParameter
        case Parameter
          1..32767:
            PosPerRot := Parameter
          other:
            abort @InvalidParameter
      else
        OutputNum(PosPerRot, 4)
    else                                                ' Only 16 elseifs, nest within an else for more
      if strcomp(@InputBuffer, string("KI"))              ' Command: Set
        if Parameter := OptionalParameter
          Parameter := ParseNum(Parameter)
          CheckLastParameter
          case Parameter
            0..32767:
              Ki := Parameter
            other:
              abort @InvalidParameter
        else
          OutputNum(Ki, 4)
      elseif strcomp(@InputBuffer, string("STORE"))         ' 1 Command: 
        Parameter := NextParameter
        CheckLastParameter
        if     strcomp(Parameter, string("HOLD"))
          WriteEEPROM(@Hold, @Hold + 3, @Hold)
        elseif strcomp(Parameter, string("PACE"))
          WriteEEPROM(@Pace, @Pace, @Pace)
        elseif strcomp(Parameter, string("PPR"))
          WriteEEPROM(@PosPerRot, @PosPerRot + 3, @PosPerRot)
        elseif strcomp(Parameter, string("SCALE"))  
          WriteEEPROM(@PulseScale, @PulseScale + 3, @PulseScale)
        elseif strcomp(Parameter, string("BAUD"))
          WriteEEPROM(@Baudrate, @Baudrate + 3, @Baudrate)
        elseif strcomp(Parameter, string("TXPIN"))
          WriteEEPROM(@Tx, @Tx, @Tx)
          WriteEEPROM(@Baudmode, @Baudmode, @Baudmode)
        elseif strcomp(Parameter, string("RXPIN"))
          WriteEEPROM(@Rx, @Rx, @Rx)
          WriteEEPROM(@Baudmode, @Baudmode, @Baudmode)
        elseif strcomp(Parameter, string("DZ"))
          WriteEEPROM(@Deadzone, @Deadzone, @Deadzone)
        elseif strcomp(Parameter, string("LZ"))
          WriteEEPROM(@KiLiveZone, @KiLiveZone, @KiLiveZone)
        elseif strcomp(Parameter, string("RAMP"))
          WriteEEPROM(@MaxPowAccel, @MaxPowAccel + 3, @MaxPowAccel)
        elseif strcomp(Parameter, string("ACC"))
          WriteEEPROM(@MaxPosAccel, @MaxPosAccel + 3, @MaxPosAccel)
        elseif strcomp(Parameter, string("KP"))
          WriteEEPROM(@Kp, @Kp + 3, @Kp)
        elseif strcomp(Parameter, string("KI"))
          WriteEEPROM(@Ki, @Ki + 3, @Ki)
        elseif strcomp(Parameter, string("KIMAX"))
          WriteEEPROM(@KiMax, @KiMax + 3, @KiMax)
        elseif strcomp(Parameter, string("KIT"))
          WriteEEPROM(@KiTimeDampen, @KiTimeDampen + 3, @KiTimeDampen)
        elseif strcomp(Parameter, string("KIP"))
          WriteEEPROM(@KiLimitByPwr, @KiLimitByPwr + 3, @KiLimitByPwr)
        else
          abort @InvalidParameter
      elseif strcomp(@InputBuffer, string("RESTORE"))    ' 2 Command:  
        CheckLastParameter
        Restore
      elseif strcomp(@InputBuffer, string("PACE"))       ' 3 Command: 
        if Parameter := OptionalParameter
          Parameter := ParseNum(Parameter)
          CheckLastParameter
          if 0 =< Parameter and Parameter =< 250
            Pace := Parameter
          else
            abort @InvalidParameter
        else
          OutputNum(Pace, 4)
      elseif strcomp(@InputBuffer, string("STATUS"))           ' 4 Command: Respond with version number
        CheckLastParameter
        OutputNum(PDRunning, 4)
      elseif strcomp(@InputBuffer, string("REBOOT"))           ' 5 Command: REBOOT board.
        CheckLastParameter
        REBOOT
      else
        abort @InvalidCommand

  return 0


Pri InterpolateMidVariables | Difference
'' Sets MidVelocity and MidPosition variables to values that would create the current SetPower at the current MotorPosition  

  ' Set the left MidPosition to the current motor position
  if Difference := SetPower[LEFT] / Kp                  ' Determine left MidPosition to MotorPosition offset
    if Difference => DEADZONE                           ' Adjust for the deadzone   
      Difference -= DEADZONE 
    elseif Difference =< -DEADZONE
      Difference += DEADZONE
    MidPosition[LEFT] := MotorPosition[LEFT] + Difference ' Add it back to the current Motor Position
  else 
    MidPosition[LEFT] := MotorPosition[LEFT]
  MidPosAcc[LEFT]~                                      ' Clear the fractional part
   
  ' Do the same for the right MidPosition
  if Difference := SetPower[RIGHT] / Kp
    if Difference => DEADZONE
      Difference -= DEADZONE
    elseif Difference =< -DEADZONE
      Difference += DEADZONE
    MidPosition[RIGHT] := MotorPosition[RIGHT] + Difference
  else
    MidPosition[RIGHT] := MotorPosition[RIGHT]
  MidPosAcc[RIGHT]~
   
  ' Set the left MidVelocity to the current motor speed
  MidVelocity[LEFT] := MotorSpeed[LEFT]
  MidVelAcc[LEFT]~                                      ' Clear the fractional part
   
  ' Do the same for the right MidVelocity
  MidVelocity[RIGHT] := MotorSpeed[RIGHT]
  MidVelAcc[LEFT]~

  ' Clear Ki accumulators
  KiAcc[LEFT]~
  KiAcc[RIGHT]~


Pri PDLoop | NextCNT                                    '' Measure, set, and maintain wheel position

  
  Encoders[LEFT].Start(L_ENCODERS_PIN, 1, 2, @MotorPosition[LEFT])                  
  Encoders[RIGHT].Start(R_ENCODERS_PIN, 1, 2, @MotorPosition[RIGHT])                
  Motors.Start                                          ' Start the motor controller
  
  outa[L_FAULT_LED]~~
  outa[R_FAULT_LED]~~
  NextCNT := cnt                                        ' Set up a timed loop
  repeat
    waitcnt(NextCNT += constant(_clkfreq / DIVISOR / 2))' Loop DIVSOR times per second (second half)
    Motors.Left(PDIteration(LEFT))                      ' Service left wheel and motor
    if !ina[Motors#MOTOR_L_FAULT] & 1                   ' Check for motor fault and: Light fault LED
      LeftFaultCnt := LeftFaultCnt + 1 <# NUM_HB_FAULTS_TO_STOP '   Increment (bounded) consecutive fault counter
      if LeftFaultCnt => NUM_HB_FAULTS_TO_STOP            '   If number of faults exceeds threshold then
        Motors.Stop                                       '   Force off all motor power
        dira[L_FAULT_LED]~~                               '   Turn on fault LED
        PDRunning := -2                                   '   Let parser know motor power is disabled
        waitpeq(0,1,0)                                    '   Stop servicing the motors
    else
      LeftFaultCnt~                                       '   No fault. Clear consecutive fault counter.

    waitcnt(NextCNT += constant(_clkfreq / DIVISOR / 2))' Do the same for the right motor
    Motors.Right(PDIteration(RIGHT))
    if !ina[Motors#MOTOR_R_FAULT] & 1
      RightFaultCnt := RightFaultCnt + 1 <# NUM_HB_FAULTS_TO_STOP
      if RightFaultCnt => NUM_HB_FAULTS_TO_STOP
        Motors.Stop
        dira[R_FAULT_LED]~~
        PDRunning := -3
        waitpeq(0,1,0)
    else
      RightFaultCnt~

  
Pri PDIteration(Side) | MotorPositionSample, Difference, Limit, KiPower
'' Read the wheel's speed and position, and set its power

  MotorPositionSample := MotorPosition[Side]                ' Sample at the beginning to remove jitter
  ' Compare current position to position from 1/2 second ago to determine speed in positions per second
  If Side == LEFT
    PrvMotPossIdx++
    PrvMotPossIdx //= HALF_SEC
  MotorSpeed[Side] := MotorPositionSample - PrevMotPoss[Side * HALF_SEC + PrvMotPossIdx]
  PrevMotPoss[Side * HALF_SEC + PrvMotPossIdx] := MotorPositionSample ' Record current position for next round

  if ||MotorSpeed[Side] =< 1                            ' Keep track of how long the motor hasn't been moving
    StillCnt[Side]++
    StillCnt[Side] <#= DIVISOR
  else
    StillCnt[Side]~

  ' Keep the mid point from traveling too far away from the current motor position
      
  Difference := MidPosition[Side] - MotorPositionSample
  ' Check to see if offset is higher than it should be for the current power level (When motor is moving)
  if ||Difference > SAFEZONE and ||Difference > ||SetPower[Side] / Kp + DEADZONE
    ' If so, bring the set point closer to the physical position
    ' MidPosition equals MotorPositionSample approaching MidPosition, as limited by SetPower / Kp + DEADZONE
    MidPosition[Side] := MotorPositionSample - (-(SetPower[Side] / Kp + DEADZONE) #> -Difference <# (SetPower[Side] / Kp + DEADZONE))

    if StillCnt[Side] => DIVISOR                        ' If the offset was high and the motor hasn't moved for the last second
      Motors.Stop                                       ' Then something is wrong; force off all motor power
      PDRunning := -4 - Side                            ' Let parser know motor power is disabled
      Blink(24, |< L_FAULT_LED | |< R_FAULT_LED)        ' Stop servicing the motors

  ' Accumulate Ki
  'if || Difference => DEADZONE
  if Difference
    KiAcc[Side] := -KiMax #> KiAcc[Side] * KiTimeDampen / 100 + (MidPosition[Side] - MotorPositionSample) * Ki <# KiMax
  else
    KiAcc[Side]~

  case Mode                                             ' Set motor power based on current control method
    POWER:                                              ' Run the motors at a set power level
      ' MotorPower approaches SetPower as limited by MaxPowAccel
      return MotorPower[Side] += -MaxPowAccel #> (SetPower[Side] - MotorPower[Side]) <# MaxPowAccel
      StillCnt[Side]~

    VELOCITY:                                           ' Maintain the motors at a set velocity
      ' MidVelocity / HALF_SEC approaches SetVelocity as limited by MaxPosAccel
      MidVelAcc[Side] += -MaxPosAccel #> (SetVelocity[Side] - MidVelocity[Side]) * HALF_SEC - MidVelAcc[Side] <# MaxPosAccel
      MidVelocity[Side] += MidVelAcc[Side] / HALF_SEC
      MidVelAcc[Side] //= HALF_SEC

      ' MidPosition / HALF_SEC increases by MidVelocity / HALF_SEC
      MidPosAcc[Side] += MidVelocity[Side]                             
      MidPosition[Side] += MidPosAcc[Side] / HALF_SEC
      MidPosAcc[Side] //= HALF_SEC

      ' Measure motors physical distance from the set point
      Difference := MidPosition[Side] - MotorPositionSample
      if Difference => DEADZONE                         ' Adjust for the deadzone   
        Difference -= DEADZONE
      elseif Difference =< -DEADZONE
        Difference += DEADZONE
      else
        Difference~

      ' Limit KiPower
      KiPower := KiLiveZone + (0 #> Motors#MAX_ON_TIME - (MotorPower[Side] * KiLimitByPwr / 256))   
      'KiPower := 0 #> Motors#MAX_ON_TIME * 10 - ||(MotorPower[Side] * KiLimitByPwr) + KiLiveZone  

      ' SetPower is proportional to the motors physical distance from the set point, limited by Motors#MAX_ON_TIME
      SetPower[Side] := -Motors#MAX_ON_TIME #> Difference * Kp + (-KiPower #> KiAcc[Side] <# KiPower) <# Motors#MAX_ON_TIME
      'SetPower[Side] := -Motors#MAX_ON_TIME #> Difference * Kp <# Motors#MAX_ON_TIME

      ' MotorPower approaches SetPower as limited by MaxPowAccel
      return MotorPower[Side] += -MaxPowAccel #> (SetPower[Side] - MotorPower[Side]) <# MaxPowAccel

    STOPPING:                                           ' Slow to a stop at a set position
      ' MidVelocity / HALF_SEC approaches SetVelocity as limited by MaxPosAccel / DIVISOR
      MidVelAcc[Side] += -MaxPosAccel #> (SetVelocity[Side] - MidVelocity[Side]) * HALF_SEC - MidVelAcc[Side] <# MaxPosAccel
      MidVelocity[Side] += MidVelAcc[Side] / HALF_SEC
      MidVelAcc[Side] //= HALF_SEC
         
      ' MidPosition / HALF_SEC approaches SetPosition as limited by the deceleration curve
      Limit := ^^||(constant(8 * HALF_SEC * HALF_SEC) * (SetPosition[Side] - MidPosition[Side])/ Decel) * Decel / 2
      MidPosAcc[Side] += -Limit #> (SetPosition[Side] - MidPosition[Side]) * constant(HALF_SEC * HALF_SEC) + MidPosAcc[Side] <# Limit
      MidPosition[Side] += MidPosAcc[Side] / constant(HALF_SEC * HALF_SEC)
      MidPosAcc[Side] //= constant(HALF_SEC * HALF_SEC)

      ' Measure motors physical distance from the set point
      Difference := MidPosition[Side] - MotorPositionSample

      if Difference => DEADZONE                         ' Adjust for the deadzone   
        Difference -= DEADZONE
      elseif Difference =< -DEADZONE
        Difference += DEADZONE
      else
        Difference~

      ' Limit KiPower
      KiPower := KiLiveZone + (0 #> Motors#MAX_ON_TIME - (MotorPower[Side] * KiLimitByPwr / 256))   
      'KiPower := 0 #> Motors#MAX_ON_TIME * 10 - ||(MotorPower[Side] * KiLimitByPwr) + KiLiveZone  

      ' SetPower is proportional to the motors physical distance from the set point, limited by Motors#MAX_ON_TIME
      SetPower[Side] := -Motors#MAX_ON_TIME #> Difference * Kp + (-KiPower #> KiAcc[Side] <# KiPower) <# Motors#MAX_ON_TIME

      ' MotorPower approaches SetPower as limited by MaxPowAccel
      return MotorPower[Side] += -MaxPowAccel #> (SetPower[Side] - MotorPower[Side]) <# MaxPowAccel
   
    POSITION:                                           ' Turn along an arc
      if InsideWheel == Side
        ' MidVelocity / HALF_SEC approaches SetVelocity as limited by MaxPosAccel / DIVISOR
        MidVelAcc[Side] += -MaxPosAccel #> (SetVelocity[Side] - MidVelocity[Side]) * HALF_SEC - MidVelAcc[Side] <# MaxPosAccel
        MidVelocity[Side] += MidVelAcc[Side] / HALF_SEC
        MidVelAcc[Side] //= HALF_SEC
         
        ' MidPosition proportionally follows MidPosition of other side
        MidPosition[Side] := StartPosition[Side] + (MidPosition[Side ^ 1] - StartPosition[Side ^ 1]) * (SetPosition[Side] - StartPosition[Side]) / (SetPosition[Side ^ 1] - StartPosition[Side ^ 1])
                              
      else
        ' MidVelocity / HALF_SEC approaches SetVelocity as limited by MaxPosAccel / DIVISOR
        MidVelAcc[Side] += -MaxPosAccel #> (SetVelocity[Side] - MidVelocity[Side]) * HALF_SEC - MidVelAcc[Side] <# MaxPosAccel
        MidVelocity[Side] += MidVelAcc[Side] / HALF_SEC
        MidVelAcc[Side] //= HALF_SEC
         
        ' MidPosition approaches SetPosition, at MidVelocity, as limited by the deceleration curve
        Limit := MidVelocity[Side] * HALF_SEC + MidVelAcc[Side] <# ^^||(constant(8 * HALF_SEC * HALF_SEC) * (SetPosition[Side] - MidPosition[Side])/ MaxPosAccel) * MaxPosAccel / 2
        MidPosAcc[Side] += -Limit #> (SetPosition[Side] - MidPosition[Side]) * constant(HALF_SEC * HALF_SEC) + MidPosAcc[Side] <# Limit
        MidPosition[Side] += MidPosAcc[Side] / constant(HALF_SEC * HALF_SEC)
        MidPosAcc[Side] //= constant(HALF_SEC * HALF_SEC)
         
      ' Measure motors physical distance from the set point
      Difference := MidPosition[Side] - MotorPositionSample
       
      if Difference => DEADZONE                           ' Adjust for the deadzone   
        Difference -= DEADZONE
      elseif Difference =< -DEADZONE
        Difference += DEADZONE
      else
        Difference~
       
      ' Limit KiPower
      KiPower := KiLiveZone + (0 #> Motors#MAX_ON_TIME - (MotorPower[Side] * KiLimitByPwr / 256))   
      'KiPower := 0 #> Motors#MAX_ON_TIME * 10 - ||(MotorPower[Side] * KiLimitByPwr) + KiLiveZone    
       
      ' SetPower is proportional to the motors physical distance from the set point, limited by Motors#MAX_ON_TIME
      SetPower[Side] := -Motors#MAX_ON_TIME #> Difference * Kp + (-KiPower #> KiAcc[Side] <# KiPower) <# Motors#MAX_ON_TIME
       
      ' MotorPower approaches SetPower as limited by MaxPowAccel
      return MotorPower[Side] += -MaxPowAccel #> (SetPower[Side] - MotorPower[Side]) <# MaxPowAccel

    other:                                              ' Invalid state
      return MotorPower[Side]~                          ' Stop the motor

     
Pri CheckShortedAUX

  dira[AUX2]~~                                          ' Drive AUX2 low
  waitcnt(MILISECOND + cnt)                             ' Wait 1 ms to ignore capative coupling
  if ina[AUX2..AUX1] <> %00                             ' Check for AUX2 and AUX1 low
    dira[AUX2]~
    return
  outa[AUX2]~~                                          ' Drive AUX2 high
  waitcnt(MILISECOND + cnt)                             ' Wait 1 ms to ignore capative coupling
  if ina[AUX2..AUX1] <> %11                             ' Check for AUX2 and AUX1 high
    outa[AUX2]~
    dira[AUX2]~
    return

  outa[AUX2]~                                           ' Float AUX2
  dira[AUX2]~
  waitcnt(MILISECOND + cnt)                             ' Wait 1 ms to ignore capative coupling
  if ina[AUX2..AUX1] <> %11                             ' Check for AUX2 and AUX1 floating high
    return

  dira[AUX1]~~                                          ' Drive AUX1 low
  waitcnt(MILISECOND + cnt)                             ' Wait 1 ms to ignore capative coupling
  if ina[AUX2..AUX1] <> %00                             ' Check for AUX2 and AUX1 low
    outa[AUX1]~
    return
  outa[AUX1]~~                                          ' Drive AUX1 high
  waitcnt(MILISECOND + cnt)                             ' Wait 1 ms to ignore capative coupling
  if ina[AUX2..AUX1] <> %11                             ' Check for AUX2 and AUX1 high
    outa[AUX1]~
    dira[AUX1]~
    return

  outa[AUX1]~                                           ' Float AUX1
  dira[AUX1]~

  Restore

  outa[L_FAULT_LED]~~
  outa[R_FAULT_LED]~~
  repeat 4
    dira[L_FAULT_LED]~~
    waitcnt(constant(_clkfreq / 4) + cnt)
    dira[L_FAULT_LED]~
    dira[R_FAULT_LED]~~
    waitcnt(constant(_clkfreq / 4) + cnt)
    dira[R_FAULT_LED]~
  outa[L_FAULT_LED]~
  outa[R_FAULT_LED]~


Pri Restore

  Hold := DEFAULT_HOLD
  Pace := DEFAULT_PACE
  PosPerRot := DEFAULT_POSITIONS_PER_ROTATION
  PulseScale := DEFAULT_PULSE_SCALE
  Baudrate := DEFAULT_BAUDRATE
  Baudmode := DEFAULT_BAUDMODE
  Tx := DEFAULT_TX
  Rx := DEFAULT_RX
  Deadzone := DEFAULT_DEADZONE
  KiLiveZone := DEFAULT_KI_LIVE_ZONE
  MaxPowAccel := DEFAULT_MAX_POW_ACC
  MaxPosAccel := DEFAULT_MAX_POS_ACC
  Kp := DEFAULT_KP
  Ki := DEFAULT_KI
  KiMax := DEFAULT_KI_MAX
  KiTimeDampen := DEFAULT_KI_TIME_DAMPEN
  KiLimitByPwr := DEFAULT_KI_LIMIT_BY_POWER

  WriteEEPROM(@Hold, @Hold + 3, @Hold)
  WriteEEPROM(@Pace, @Pace, @Pace)
  WriteEEPROM(@PosPerRot, @PosPerRot + 3, @PosPerRot)
  WriteEEPROM(@PulseScale, @PulseScale + 3, @PulseScale)
  WriteEEPROM(@Baudrate, @Baudrate + 3, @Baudrate)
  WriteEEPROM(@Baudmode, @Baudmode, @Baudmode)
  WriteEEPROM(@Tx, @Tx, @Tx)
  WriteEEPROM(@Rx, @Rx, @Rx)
  WriteEEPROM(@Deadzone, @Deadzone, @Deadzone)
  WriteEEPROM(@KiLiveZone, @KiLiveZone, @KiLiveZone)
  WriteEEPROM(@MaxPowAccel, @MaxPowAccel + 3, @MaxPowAccel)
  WriteEEPROM(@MaxPosAccel, @MaxPosAccel + 3, @MaxPosAccel)
  WriteEEPROM(@Kp, @Kp + 3, @Kp)
  WriteEEPROM(@Ki, @Ki + 3, @Ki)
  WriteEEPROM(@KiMax, @KiMax + 3, @KiMax)
  WriteEEPROM(@KiTimeDampen, @KiTimeDampen + 3, @KiTimeDampen)
  WriteEEPROM(@KiLimitByPwr, @KiLimitByPwr + 3, @KiLimitByPwr)
   

Pri NextParameter                                       '' Condition the next input parameter and return its pointer

  repeat until ++ParseIndex => InputIndex               ' Ignore whitespace
    case InputBuffer[ParseIndex]                        ' First character is always whitespace
      0, HT, " ", ",":
      other:
        quit                                            ' ParseIndex points to first non-whitespace character
  if ParseIndex => InputIndex                           ' If at the end of the buffer (or passed it, just in case)
    abort @TooFewParameters                             '  then there are no more parameters
  result := @InputBuffer[ParseIndex]                    ' When responding, point to the next parameter,

  repeat ParseIndex from ParseIndex to InputIndex - 1   ' But first...
    case InputBuffer[ParseIndex]                         
      NUL, HT, " ", "," :                               '  Null terminate the parameter           
        InputBuffer[ParseIndex]~
        quit
      "a".."z" :                                        '  Set all command characters to uppercase
        InputBuffer[ParseIndex] -= constant("a" - "A")


Pri OptionalParameter                                   '' Condition the next input parameter and return its pointer

  repeat until ++ParseIndex => InputIndex               ' Ignore whitespace
    case InputBuffer[ParseIndex]                        ' First character is always whitespace
      0, HT, " ", ",":
      other:
        quit                                            ' ParseIndex points to first non-whitespace character
  if ParseIndex => InputIndex                           ' If at the end of the buffer (or passed it, just in case)
    return 0                                            '  then there are no more parameters
  result := @InputBuffer[ParseIndex]                    ' When responding, point to the next parameter,

  repeat ParseIndex from ParseIndex to InputIndex - 1   ' But first...
    case InputBuffer[ParseIndex]                         
      NUL, HT, " ", "," :                               '  Null terminate the parameter           
        InputBuffer[ParseIndex]~
        quit
      "a".."z" :                                        '  Set all command characters to uppercase
        InputBuffer[ParseIndex] -= constant("a" - "A")


Pri CheckLastParameter                                  '' Abort if there are any unparsed input parameters

  repeat until (InputBuffer[ParseIndex] <> 0 and InputBuffer[ParseIndex] <> " " and InputBuffer[ParseIndex] <> "," and InputBuffer[ParseIndex] <> HT) or ParseIndex == InputIndex - 1
    ++ParseIndex

  ifnot ParseIndex == InputIndex - 1                    '   Ensure that there are no further arguments           
    abort @TooManyParameters                            '   To prevent changes with a "Too many parameters" error


Pri CheckHwVer

  if HWVer <> EXPECTED_HW_VER  
    abort @IncorrectFirmware


Pri ParseNum(Pointer) | Character, Negate               '' Interpret an ASCII hexadecimal string

  'result~                                              ' This occurs automatically when method starts
  case Base
    DECIMAL:
      Negate := 1
      if byte[Pointer] == "-"                           ' Set negate to -1 if input is negative
        Negate~~
        Pointer++

      repeat 10                                         ' The largest storable number, 2 ^ 31, is ten digits long
        case Character := byte[Pointer++]
          NUL :                                         ' Look for the end of the buffer
            Pointer--                                   ' Keep the pointer at the end of the buffer, not after
            return result * Negate                      ' Apply sign
          "0".."9" :                                    ' Only accept valid numerals
            result := result * 10 + Character - "0"
            if result < 0                               ' If the result is negative, there was an overflow, so abort
              abort @InvalidParameter
          other :                                       ' Error out on non-numeral character
            abort @InvalidParameter

    Hex:                                                ' Interpret an ASCII hexadecimal string
      repeat 4
        case Character := byte[Pointer++]
          NUL :
            Pointer--
            return ~~result
          "0".."9" :
            result := result << 4 + Character - "0"
          "A".."F" :
            result := result << 4 + Character - constant("A" - 10)
          "a".."f" :
            result := result << 4 + Character - constant("a" - 10)
          other :
            abort @InvalidParameter
      if byte[Pointer]                                    ' Make sure there are no remaining characters after parsing first four.                         
        abort @InvalidParameter


Pri OutputChr(Char)                                     '' Add a character to the output buffer

  if OutputIndex < BUFFER_LENGTH - 3
    OutputBuffer[OutputIndex++] := Char
    OutputBuffer[OutputIndex]~

  
Pri OutputStr(Pointer)                                  '' Concatenate a string to the end of the output buffer

  if strsize(Pointer) + 1 =< BUFFER_LENGTH - OutputIndex                        ' Check for overflow
    bytemove(@OutputBuffer + OutputIndex, Pointer, strsize(Pointer) + 1)        ' Copy the string to the buffer, including the terminator
    OutputIndex += strsize(Pointer)                                             ' Increment the output buffer index by the length of the sting, not including the terminator


Pri OutputNum(Value, HexDigits) | Index, WasNegX, NonZero '' Create a hexadecimal string and concatenate the end of the output buffer

  case Base
    HEX:                                                ' Print a hexadecimal number
      if OutputIndex < BUFFER_LENGTH - HexDigits - 2
        Value <<= (8 - HexDigits) << 2
        repeat HexDigits
          OutputBuffer[OutputIndex++] := lookupz((Value <-= 4) & $F : "0".."9", "A".."F")
        OutputBuffer[OutputIndex]~
     
    DECIMAL:                                            ' Print a decimal number
      WasNegX := Value == NEGX                          ' Check for max negative
      if Value < 0
        Value := ||(Value + WasNegX)                    ' If negative, make positive; adjust for max negative
        OutputChr("-")                                  ' and output sign
     
      Index := 1_000_000_000                            ' Initialize divisor
     
      NonZero~
      repeat 10                                         ' Loop for 10 digits
        if Value => Index                                                       
          OutputChr(Value / Index + "0" + WasNegX * (Index == 1)) ' If non-zero digit, output digit; adjust for max negative
          Value //= Index                               ' and digit from value
          NonZero~~                                     'flag non-zero found
        elseif NonZero or Index == 1
          OutputChr("0")                                ' If zero digit (or only digit) output it
        Index /= 10                                     ' Update divisor
   

Pri SendResponse | Index, Split, stringptr              '' Transmit the string in the output buffer and clear the buffer

  Split~
  if Pace    
    waitcnt(MILISECOND * Pace + cnt)
    repeat Index from 0 to OutputIndex - 1                                                                              
      if OutputBuffer[Index] == " "
        OutputBuffer[Index]~
        Split := Index
        quit
    
  Term.Str(@OutputBuffer)                               ' Transmit the buffer contents
  if Split
    Term.tx(" ")
    waitcnt(MILISECOND * Pace + cnt)
    Term.Str(@OutputBuffer + Split + 1)
  Term.Str(@Prompt)                                     ' Transmit the prompt
  InputIndex~                                           ' Clear the buffers
  OutputBuffer~
  OutputIndex~


Pri Blink(CtrBit, PinMask) | NextCnt

  outa := PinMask
  |< CtrBit
  NextCnt := CtrBit + cnt  
  repeat
    waitcnt(NextCnt += CtrBit)
    dira ^= PinMask
        
  
Pri WriteEEPROM(startAddr, endAddr, eeStart) | addr, page, eeAddr

  ''Copy startAddr..endAddr from main RAM to EEPROM beginning at eeStart address.

  addr := startAddr                              ' Initialize main RAM index
  eeAddr := eeStart                              ' Initialize EEPROM index
  repeat
    page := addr+64-eeAddr//64<#endaddr+1        ' Find next EEPROM page boundary
    SetAddr(eeAddr)                              ' Give EEPROM starting address
    repeat                                       ' Bytes -> EEPROM until page boundary
      SendByte(byte[addr++])
    until addr == page
    i2cstop                                      ' From 24LC256's page buffer -> EEPROM
    eeaddr := addr - startAddr + eeStart         ' Next EEPROM starting address
  until addr > endAddr                           ' Quit when RAM index > end address


Pri ReadEEPROM(startAddr, endAddr, eeStart) | addr

  ''Copy from EEPROM beginning at eeStart address to startAddr..endAddr in main RAM.
  
  SetAddr(eeStart)                               ' Set EEPROM's address pointer 
  i2cstart
  SendByte(%10100001)                            ' EEPROM I2C address + read operation
  if startAddr == endAddr
    addr := startAddr
  else
    repeat addr from startAddr to endAddr - 1      ' Main RAM index startAddr to endAddr
      byte[addr] := GetByte                        ' GetByte byte from EEPROM & copy to RAM 
      SendAck(I2C_ACK)                             ' Acknowledge byte received
  byte[addr] := GetByte                            ' GetByte byte from EEPROM & copy to RAM 
  SendAck(I2C_NACK)
  i2cstop                                        ' Stop sequential read


Pri SetAddr(addr)

  'Sets EEPROM internal address pointer.

  poll                                           ' Poll until EEPROM acknowledges
  SendByte(addr >> 8)                            ' Send address high byte
  SendByte(addr)                                 ' Send address low byte


Pri Poll | ackbit

  ' Poll until acknowledge.  This is especially important if the 24LC256 is copying from
  ' buffer to EEPROM.

  ackbit~~                                       ' Make acknowledge 1
  repeat                                         ' Send/check acknowledge loop
    i2cstart                                     ' Send I2C start condition
    ackbit := SendByte(%10100000)                ' Write command with EEPROM's address
  while ackbit                                   ' Repeat while acknowledge is not 0

  
Pri I2cStart

  ' I2C start condition.  SDA transitions from high to low while the clock is high.
  ' SCL does not have the pullup resistor called for in the I2C protocol, so it has to be
  ' set high. (It can't just be set to inSendByte because the resistor won't pull it up.)

  outa[SCL]~~                                    ' SCL pin outSendByte-high
  dira[SCL]~~
  dira[SDA]~                                     ' Let pulled up SDA pin go high
  outa[SDA]~                                     ' Transition SDA pin low
  dira[SDA]~~                                    ' SDA -> outSendByte for SendByte method


Pri SendByte(b) : ackbit | i

  ' Shift a byte to EEPROM msb first.  Return if EEPROM acknowledged.  Returns
  ' acknowledge bit.  0 = ACK, 1 = NACK.

  b ><= 8                                        ' Reverse bits for shifting msb right
  outa[SCL]~                                     ' SCL low, SDA can change
  repeat 8                                       ' 8 reps sends 8 bits
    outa[SDA] := b                               ' Lowest bit sets state of SDA
    outa[SCL]~~                                  ' Pulse the SCL line
    outa[SCL]~
    b >>= 1                                      ' Shift b right for next bit
  ackbit := GetAck                               ' Call GetByteAck and return EEPROM's Ack


Pri GetAck : ackbit

  ' GetByte and return acknowledge bit transmitted by EEPROM after it receives a byte.
  ' 0 = ACK, 1 = NACK.

  dira[SDA]~                                     ' SDA -> SendByte so 24LC256 controls
  outa[SCL]~~                                    ' Start a pulse on SCL
  ackbit := ina[SDA]                             ' GetByte the SDA state from 24LC256
  outa[SCL]~                                     ' Finish SCL pulse
  outa[SDA]~                                     ' SDA will hold low
  dira[SDA]~~                                    ' SDA -> outSendByte, master controls

  
Pri I2cStop

  ' Send I2C stop condition.  SCL must be high as SDA transitions from low to high.
  ' See note in i2cStart about SCL line.

  outa[SDA]~                                     ' SDA -> outSendByte low
  dira[SDA]~~
  outa[SCL]~~                                    ' SCL -> high
  dira[SDA]~                                     ' SDA -> inSendByte GetBytes pulled up

   
Pri GetByte : value

  ' Shift in a byte msb first.  

  value~                                         ' Clear value
  dira[SDA]~                                     ' SDA input so 24LC256 can control
  repeat 8                                       ' Repeat shift in eight times
    outa[SCL]~~                                  ' Start an SCL pulse
    value <<= 1                                  ' Shift the value left
    value += ina[SDA]                            ' Add the next most significant bit
    outa[SCL]~                                   ' Finish the SCL pulse


Pri SendAck(ackbit)

  ' Transmit an acknowledgement bit (ackbit).

  outa[SDA]:= ackbit                             ' Set SDA output state to ackbit
  dira[SDA]~~                                    ' Make sure SDA is an output
  outa[SCL]~~                                    ' Send a pulse on SCL
  outa[SCL]~
  dira[SDA]~                                     ' Let go of SDA


Dat

Prompt                  byte CR, 0, 0
Nack                    byte "ERROR ", 0
Overflow                byte "1", 0, "- Overflow", 0
InvalidCommand          byte "2", 0, "- Invalid command", 0
InvalidParameter        byte "3", 0, "- Invalid parameter", 0
TooFewParameters        byte "4", 0, "- Too few parameters", 0
TooManyParameters       byte "5", 0, "- Too many parameters", 0
MotorError              byte "6", 0, "- Motor or encoder error", 0
IncorrectFirmware       byte "7", 0, "- Incorrect firmware version", 0